# This function performs one repetition of the two-times repeated 
# 5-fold stratified cross-validation on a specific data set using
# diversity forests.
#
# It takes the whole number 'iter', which corresponds to the iter-th line 
# of 'scenariogrid', which contains the necessary information
# on the iter-th setting.

evaluatesetting <- function(iter) {
  
  
  # Obtain information for the iter-th setting:

  dataid <- scenariogrid$dataid[iter]
  nmaxsplits <- scenariogrid$nmaxsplits[iter]
  proptry <- scenariogrid$proptry[iter]
  cvind <- scenariogrid$cvind[iter]
  seed <- scenariogrid$seed[iter]
  
  
  # Set seed:
  
  set.seed(seed)
  
  
  # Load data set:
  
  alldatasets <- list.files("./Online_Appendix_2/Data/Datasets")
  datasetname <- grep(paste("_id", dataid, ".Rda", sep=""), alldatasets, value=TRUE)
  
  load(paste("./Online_Appendix_2/Data/Datasets/", datasetname, sep=""))
  
  
  # Make CV division:
  
  Kcv <- 5
  
  cvdiv <- makeCVdiv(data=dataset, yname="Ytarget", ncv=Kcv)
  
  
  # Perform cross-validation: 
  
  res <- list()
  
  for(i in 1:Kcv) {
    
    datatrain <- dataset[cvdiv!=i,]
    datatest <- dataset[cvdiv==i,]
 
    res[[i]] <- dfwrap(datatrain=datatrain, datatest=datatest, num.trees=2000, max.triedsplits=nmaxsplits, proptry=proptry)

  }
  
  
  # Return the results:
  
  return(res)
  
}





# This function performs one iteration of the two-times repeated 5-fold stratified cross-validation:

# Input parameters:

# datatrain       - training data set
# datatest        - test data set
# num.trees       - numbers of trees in the forests
# max.triedsplits - value used for nsplits
# proptry         - values used for proprtry

dfwrap <- function(datatrain, datatest, num.trees, max.triedsplits, proptry) {
  
  require("nnet")
  require("data.table")
  require("diversityForest")
  
  # Construct diversity forest:
  model <- divfor(Ytarget ~ ., data = datatrain, num.trees = num.trees, proptry=proptry, importance = "permutation", num.threads=1, verbose=FALSE, max.triedsplits=max.triedsplits)
  
  # Predict conditional class probabilities for the observations in
  # the test data set:
  predprobs <- apply(predict(model, data=datatest, predict.all=TRUE)$predictions, 1, function(x) mean(x==2))
  
  # Obtain class predictions:
  preds <- factor(ifelse(predprobs > 0.5, levels(datatest$Ytarget)[2], levels(datatest$Ytarget)[1]), levels=levels(datatest$Ytarget))
  if(any(predprobs==0.5))
    preds[predprobs==0.5] <- sample(levels(datatest$Ytarget), size=sum(predprobs==0.5), replace=TRUE)
  
  # Calculate AUC values:
  auctest <- measureAUC(probabilities=predprobs, truth=datatest$Ytarget, 
                        negative=levels(datatest$Ytarget)[1], positive=levels(datatest$Ytarget)[2])
  
  # Calculate ACC values:
  acctest <- measureACC(truth=datatest$Ytarget, response=preds)
  
  # Calculate Brier values:
  briertest <- measureBrier(probabilities=predprobs, truth=datatest$Ytarget, 
                            negative=levels(datatest$Ytarget)[1], positive=levels(datatest$Ytarget)[2])
  
  # Return results:
  return(list(acctest=acctest, auctest=auctest, briertest=briertest,
              varimp=model$variable.importance))
  
}





# Function calculating AUC values (taken from "mlr" R package):

measureAUC <- function (probabilities, truth, negative, positive) 
{
  if (is.factor(truth)) {
    i = as.integer(truth) == which(levels(truth) == positive)
  }
  else {
    i = truth == positive
  }
  if (length(unique(i)) < 2L) {
    stop("truth vector must have at least two classes")
  }
  if (length(i) > 5000L) {
    r = frankv(probabilities)
  }
  else {
    r = rank(probabilities)
  }
  n.pos = as.numeric(sum(i))
  n.neg = length(i) - n.pos
  (sum(r[i]) - n.pos * (n.pos + 1)/2)/(n.pos * n.neg)
}



# Function calculating ACC values (taken from "mlr" R package):

measureACC <- function (truth, response) 
{
  mean(response == truth)
}



# Function calculating Brier values (taken from "mlr" R package):

measureBrier <- function (probabilities, truth, negative, positive) 
{
  y = as.numeric(truth == positive)
  mean((y - probabilities)^2)
}





# Function to generate the splittings for stratified cross-validation:

# Input parameters:

# data   - dataset
# yname  - label of the target variable
# ncv    - number of folds to use

makeCVdiv <- function(data, yname="y", ncv=5) {
  
  cvid <- rep(NA, length = nrow(data))
  for(i in levels(data[, yname])) cvid[data[, yname] == i] <- sample(rep(1:ncv, length = sum(data[, yname] == i)))
  
  cvid
  
}
