# This function performs one repetition of the two-times repeated 
# 5-fold stratified cross-validation on a specific data set using
# one of the four compared methods.
#
# It takes the whole number 'iter', which corresponds to the iter-th line 
# of 'scenariogrid', which contains the necessary information
# on the iter-th setting.

evaluatesetting <- function(iter) {
  
  # Obtain information for the iter-th setting:

  dataid <- scenariogrid$dataid[iter]
  method <- scenariogrid$method[iter]
  cvind <- scenariogrid$cvind[iter]
  seed <- scenariogrid$seed[iter]
  
  num_trees_pre <- 1000
  num_trees <- 2000
  
  
  # Set seed:
  
  set.seed(seed)
  
  
  # Load data set:
  
  alldatasets <- list.files("./Online_Appendix_2/Data/Datasets")
  datasetname <- grep(paste("_id", dataid, ".Rda", sep=""), alldatasets, value=TRUE)
  
  load(paste("./Online_Appendix_2/Data/Datasets/", datasetname, sep=""))
  
  
  # Make CV division:
  
  Kcv <- 5
  
  cvdiv <- makeCVdiv(data=dataset, yname="Ytarget", ncv=Kcv)
    
  
  # Perform cross-validation: 
  
  res <- list()
  
  for(i in 1:Kcv) {
    
    datatrain <- dataset[cvdiv!=i,]
    datatest <- dataset[cvdiv==i,]
 
    if(method=="rf") {
      
      res[[i]] <- rfwrap(datatrain=datatrain, datatest=datatest, num.trees.pre=num_trees_pre, num.trees=num_trees)
      
    }
    
    if(method=="rfextratrees1") {
      
      res[[i]] <- rfextratreeswrap(datatrain=datatrain, datatest=datatest, num.trees.pre=num_trees_pre, num.trees=num_trees, num.random.splits = 1)
      
    }
    
    if(method=="rfextratrees5") {
      
      res[[i]] <- rfextratreeswrap(datatrain=datatrain, datatest=datatest, num.trees.pre=num_trees_pre, num.trees=num_trees, num.random.splits = 5)
      
    }
    
    if(method=="df") {
      
      res[[i]] <- dfwrap(datatrain=datatrain, datatest=datatest, num.trees.pre=num_trees_pre, num.trees=num_trees)
      
    }
 
  }
  
  
  # Return the results:
  
  return(res)
  
}





# This function performs one iteration of the two-times repeated 
# 5-fold stratified cross-validation using conventional random forests:

# Input parameters:

# datatrain      - training data set
# datatest       - test data set
# num.trees.pre  - number of trees in the forests constructed during tuning parameter optimization 
# num.trees      - number of trees in the final forest constructed after tuning parameter optimization

rfwrap <- function(datatrain, datatest, num.trees.pre, num.trees) {
  
  require("nnet")
  require("data.table")
  require("ranger")
  
  
  # Determine grid used for mtry:
  if((ncol(datatrain)-1) <= 20) {
    mtrys <- 1:(ncol(datatrain)-1)
  } else {
    powermtry <- log(ncol(datatrain)-1)/log(20)
    mtrys <- round((1:20)^powermtry)
  }
  
  # Perform tuning parameter optimization:
  ooberrs <- 0
  for(i in seq(along=mtrys))
    ooberrs[i] <- ranger(Ytarget ~ ., data = datatrain, num.trees = num.trees.pre, mtry=mtrys[i], write.forest=FALSE, num.threads=1, verbose=FALSE)$prediction.error
  
  mtryopt <- mtrys[which.is.max(-ooberrs)]
  
  # Construct forest:
  model <- ranger(Ytarget ~ ., data = datatrain, num.trees = num.trees, mtry=mtryopt, importance = "permutation", num.threads=1, verbose=FALSE)
  
  
  # Predict conditional class probabilities for the observations in
  # the test data set:
  predprobs <- apply(predict(model, data=datatest, predict.all=TRUE)$predictions, 1, function(x) mean(x==2))
  
  # Obtain class predictions:
  preds <- factor(ifelse(predprobs > 0.5, levels(datatest$Ytarget)[2], levels(datatest$Ytarget)[1]), levels=levels(datatest$Ytarget))
  if(any(predprobs==0.5))
    preds[predprobs==0.5] <- sample(levels(datatest$Ytarget), size=sum(predprobs==0.5), replace=TRUE)
  
  # Calculate AUC values:
  auctest <- measureAUC(probabilities=predprobs, truth=datatest$Ytarget, 
                        negative=levels(datatest$Ytarget)[1], positive=levels(datatest$Ytarget)[2])
  
  # Calculate ACC values:
  acctest <- measureACC(truth=datatest$Ytarget, response=preds)
  
  # Calculate Brier values:
  briertest <- measureBrier(probabilities=predprobs, truth=datatest$Ytarget, 
                            negative=levels(datatest$Ytarget)[1], positive=levels(datatest$Ytarget)[2])
  
  # Return results:
  return(list(acctest=acctest, auctest=auctest, briertest=briertest, paramseq=mtrys, ooberrs=ooberrs,
              paramopt=mtryopt, varimp=model$variable.importance))
  
}





# This function performs one iteration of the two-times repeated 
# 5-fold stratified cross-validation using random forests with extremely randomized trees:

# Input parameters:

# datatrain      - training data set
# datatest       - test data set
# num.trees.pre  - number of trees in the forests constructed during tuning parameter optimization 
# num.trees      - number of trees in the final forest constructed after tuning parameter optimization

rfextratreeswrap <- function(datatrain, datatest, num.trees.pre, num.trees, num.random.splits = 1) {
  
  require("nnet")
  require("data.table")
  require("ranger")
  
  
  # Determine grid used for mtry:
  if((ncol(datatrain)-1) <= 20) {
    mtrys <- 1:(ncol(datatrain)-1)
  } else {
    powermtry <- log(ncol(datatrain)-1)/log(20)
    mtrys <- round((1:20)^powermtry)
  }
  
  # Perform tuning parameter optimization:
  ooberrs <- 0
  for(i in seq(along=mtrys))
    ooberrs[i] <- ranger(Ytarget ~ ., data = datatrain, num.trees = num.trees.pre, mtry=mtrys[i], splitrule="extratrees", write.forest=FALSE, num.random.splits = num.random.splits, num.threads=1, verbose=FALSE)$prediction.error
  
  mtryopt <- mtrys[which.is.max(-ooberrs)]
    
  # Construct forest:
  model <- ranger(Ytarget ~ ., data = datatrain, num.trees = num.trees, mtry=mtryopt, importance = "permutation", splitrule="extratrees", num.random.splits = num.random.splits, num.threads=1, verbose=FALSE)
  
  
  # Predict conditional class probabilities for the observations in
  # the test data set:
  predprobs <- apply(predict(model, data=datatest, predict.all=TRUE)$predictions, 1, function(x) mean(x==2))
  
  # Obtain class predictions:
  preds <- factor(ifelse(predprobs > 0.5, levels(datatest$Ytarget)[2], levels(datatest$Ytarget)[1]), levels=levels(datatest$Ytarget))
  if(any(predprobs==0.5))
    preds[predprobs==0.5] <- sample(levels(datatest$Ytarget), size=sum(predprobs==0.5), replace=TRUE)
  
  # Calculate AUC values:
  auctest <- measureAUC(probabilities=predprobs, truth=datatest$Ytarget, 
                        negative=levels(datatest$Ytarget)[1], positive=levels(datatest$Ytarget)[2])
  
  # Calculate ACC values:
  acctest <- measureACC(truth=datatest$Ytarget, response=preds)
  
  # Calculate Brier values:
  briertest <- measureBrier(probabilities=predprobs, truth=datatest$Ytarget, 
                            negative=levels(datatest$Ytarget)[1], positive=levels(datatest$Ytarget)[2])
  
  # Return results:
  return(list(acctest=acctest, auctest=auctest, briertest=briertest, paramseq=mtrys, ooberrs=ooberrs,
              paramopt=mtryopt, varimp=model$variable.importance))
  
}





# This function performs one iteration of the two-times repeated 
# 5-fold stratified cross-validation using diversity forests:

# Input parameters:

# datatrain      - training data set
# datatest       - test data set
# num.trees.pre  - number of trees in the forests constructed during tuning parameter optimization 
# num.trees      - number of trees in the final forest constructed after tuning parameter optimization

dfwrap <- function(datatrain, datatest, num.trees.pre, num.trees) {
  
  require("nnet")
  require("data.table")
  require("diversityForest")
    
  # Perform tuning parameter optimization:
  proptrys <- c(0.05, 1)
    
  ooberrs <- 0
  for(i in seq(along=proptrys))
    ooberrs[i] <- divfor(Ytarget ~ ., data = datatrain, num.trees = num.trees.pre, proptry=proptrys[i], write.forest=FALSE, num.threads=1, verbose=FALSE, max.triedsplits=30)$prediction.error
  
  proptryopt <- proptrys[which.is.max(-ooberrs)]
  model <- divfor(Ytarget ~ ., data = datatrain, num.trees = num.trees, proptry=proptryopt, importance = "permutation", num.threads=1, verbose=FALSE, max.triedsplits=30)
  
  
  # Predict conditional class probabilities for the observations in
  # the test data set:
  predprobs <- apply(predict(model, data=datatest, predict.all=TRUE)$predictions, 1, function(x) mean(x==2))
  
  # Obtain class predictions:
  preds <- factor(ifelse(predprobs > 0.5, levels(datatest$Ytarget)[2], levels(datatest$Ytarget)[1]), levels=levels(datatest$Ytarget))
  if(any(predprobs==0.5))
    preds[predprobs==0.5] <- sample(levels(datatest$Ytarget), size=sum(predprobs==0.5), replace=TRUE)
  
  # Calculate AUC values:
  auctest <- measureAUC(probabilities=predprobs, truth=datatest$Ytarget, 
                        negative=levels(datatest$Ytarget)[1], positive=levels(datatest$Ytarget)[2])
  
  # Calculate ACC values:
  acctest <- measureACC(truth=datatest$Ytarget, response=preds)
  
  # Calculate Brier values:
  briertest <- measureBrier(probabilities=predprobs, truth=datatest$Ytarget, 
                            negative=levels(datatest$Ytarget)[1], positive=levels(datatest$Ytarget)[2])
  
  # Return results:
  return(list(acctest=acctest, auctest=auctest, briertest=briertest, ooberrs=ooberrs,
              paramopt=proptryopt, varimp=model$variable.importance))
  
}





# Function calculating AUC values (taken from "mlr" R package):

measureAUC <- function (probabilities, truth, negative, positive) 
{
  if (is.factor(truth)) {
    i = as.integer(truth) == which(levels(truth) == positive)
  }
  else {
    i = truth == positive
  }
  if (length(unique(i)) < 2L) {
    stop("truth vector must have at least two classes")
  }
  if (length(i) > 5000L) {
    r = frankv(probabilities)
  }
  else {
    r = rank(probabilities)
  }
  n.pos = as.numeric(sum(i))
  n.neg = length(i) - n.pos
  (sum(r[i]) - n.pos * (n.pos + 1)/2)/(n.pos * n.neg)
}



# Function calculating ACC values (taken from "mlr" R package):

measureACC <- function (truth, response) 
{
  mean(response == truth)
}



# Function calculating Brier values (taken from "mlr" R package):

measureBrier <- function (probabilities, truth, negative, positive) 
{
  y = as.numeric(truth == positive)
  mean((y - probabilities)^2)
}





# Function to generate the splittings for stratified cross-validation:

# Input parameters:

# data   - dataset
# yname  - label of the target variable
# ncv    - number of folds to use

makeCVdiv <- function(data, yname="y", ncv=5) {
  
  cvid <- rep(NA, length = nrow(data))
  for(i in levels(data[, yname])) cvid[data[, yname] == i] <- sample(rep(1:ncv, length = sum(data[, yname] == i)))
  
  cvid
  
}
